class Interface:
    """
    Represents *logical* graph edges, with Tube being its physical counterpart.
    """

    def __init__(self):
        ...

    def __repr__(self):
        return "Logical interface."
